from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Tuple, Iterator
from urllib.request import urlretrieve
from tempfile import NamedTemporaryFile
import warnings

from arxiv_summarizer.arxiv_wrapper import ArxivAPIWrapper
from langchain.document_loaders import ArxivLoader 
from arxiv.arxiv import Result

@dataclass(slots=True)
class ArxivPaper:
    arxiv_id: str
    name: str
    authors: List[str]
    summary: str
    published: datetime
    primary_category: str
    categories: List[str]
    pdf_url: str
    links: List[Tuple[str, str, str, str]]
    journal_ref: str
    doi: str

    # content: str
    _content: int = field(repr=False, init=False, default = None)


    @property
    def content(self, ) -> str:
        if self._content:
            return self._content

        try:
            import fitz
        except ImportError:
            raise ImportError(
                "PyMuPDF package not found, please install it with "
                "`pip install pymupdf`"
            )

        with NamedTemporaryFile(mode = "r", delete = True, prefix="arxiv_", suffix=".pdf") as arxiv_pdf:
            try:
                doc_file_name, msg = urlretrieve(self.pdf_url, arxiv_pdf.name)
                with fitz.open(doc_file_name) as doc_file:
                    self._content: str = "".join(page.get_text() for page in doc_file)
            except:
                warnings.warn(f"Cannot fetch arxiv:{self.arxiv_id}. Possibly withdrawn.")
                self._content = None
                return None

        return self._content
    
    @staticmethod
    def from_result(result: Result):
        paper = ArxivPaper(
            arxiv_id = result.entry_id.split("/")[-1],
            name = result.title,
            authors = [
                author.name for author in result.authors
            ],
            summary = result.summary,
            published = result.published,
            primary_category = result.primary_category,
            categories = result.categories,
            pdf_url = result.pdf_url,
            links = [
                (link.title, link.href, link.rel, link.content_type) for link in result.links
            ],
            journal_ref = result.journal_ref,
            doi = result.doi
        )
        return paper

def fetch_paper(query, max_docs = 1) -> Iterator[ArxivPaper]:
    """Fetches ArXiv papers based on the given query.

    Args:
        query (str): The search query for ArXiv papers.
        max_docs (int, optional): The maximum number of documents to fetch. Defaults to 1.

    Returns:
        List[ArxivPaper]: A list of ArXivPaper objects representing the fetched papers.

    Yields:
        Iterator[ArxivPaper]: A list of ArXivPaper objects representing the fetched papers.

    Note:
        The function uses the ArxivAPIWrapper to fetch paper summaries and metadata,
        and then converts the results into ArxivPaper objects.

    Raises:
        ImportError: If the PyMuPDF package (required for extracting content from PDFs)
        is not installed.

    Warnings:
        UserWarning: If the specified `max_docs` exceeds 50, it defaults to 50 documents.
        UserWarning: If a paper cannot be fetched (possibly withdrawn), a warning is issued.

    Example:
        .. code-block:: python

            from arxiv_summarizer.fetch_paper import fetch_paper

            papers = fetch_paper("machine learning", max_docs=5)
            for paper in papers:
                print(paper.name, paper.authors, paper.published)
        
    """    
    query = query.lower()

    # TODO: This check won't be necessary anymore, as the "content" 
    # is now being fetched, only when needed.
    if (max_docs > 50):
        warnings.warn("Max docs exceeded 50, defaulting to 50 documents only")

    max_docs = min(50, max_docs)

    client = ArxivAPIWrapper(
        doc_content_chars_max = None, 
        top_k_results = max_docs, 
        load_max_docs = max_docs,
        load_all_available_meta = True
    )

    results = client.get_summaries_as_docs(query = query)

    for res in results:
        yield ArxivPaper.from_result(res)



        




def chunk(content):
    '''We are currently using a manual chunking method that combines 10 sentences at one time.
        Since the model context length is around 512 we are assume that 10 sentences would create a context length of that is between 600+-100
    '''
    abst, content = clip(content)

    sentences = []
    c = 0
    k = ""
    content = content.split(". ")
    for i in range(len(content)):
        k = k + content[i] + ". "
        c = c+1
        if c == 10:
            sentences.append(k)
            c = 0
            k = ""
        elif i==len(content)-1:
            sentences.append(k)

    return sentences

def clip(content):
    ''' Using the Clip Function we are trying to clip all the contents that are above the Introduction mainly abstract, title and authors &
        all references as they are not necessary for summarization
    '''
    loc_abst = content.find("Abstract")
    loc_intro = content.find("Introduction")
    loc_refer = content.rfind("References")
    if loc_intro !=-1:
        if loc_refer !=-1:
            return content[loc_abst:loc_intro], content[loc_intro:loc_refer]
        else:
            return content[loc_abst:loc_intro], content[loc_intro:]
            print("Warning: Paper Doesn't have a References Title, may lead to overlap of references in summary")
    else:
        print("Warning: Paper Doesn't Have an Introduction Title, these may lead to overlap of summarization")

from langchain.document_loaders import ArxivLoader 

def doc_loader(search_query):
    '''The purpose of this function is to load the documents from Arxiv'''
    '''We are using the Arxiv Loader under Langchain library, which intern invokes the Arxiv API'''

    try :
        docs = ArxivLoader(query=search_query, load_max_docs=1).load()
        return docs[0].metadata, docs[0].page_content
    except Exception as e:
        print(e)

import torch

def model_selector():
    '''This function allows us to choose between 2 models, lightweight model to be used with CPU, and a heavyweight model to be used if GPU is available
    '''
    if torch.cuda.is_available():
        tokenizer_str = "facebook/bart-large-cnn"
        model_str = "facebook/bart-large-cnn"
        return tokenizer_str, model_str
    else:
        tokenizer_str = "Falconsai/text_summarization"
        model_str = "Falconsai/text_summarization"
        return tokenizer_str, model_str

def strip(content):
    ''' The purpose of this function is to strip the contents of the page of line breaks, thats helps us in better summarization, 
    as line breaks can cause the model to not function properly'''

    content = str(content)
    content = content.split("\n")
    content = " ".join(content)

    return content

from transformers import pipeline

def summarize(sent):
    ''' This is the main function that summarizes the contents of the paper. Model is initialized automatically based on CUDA availability.
    '''
    tokenizer_str, model_str = model_selector.model_selector()

    summarizer = pipeline("summarization", model=model_str, tokenizer = tokenizer_str)

    summarized = ""
    for i in sent:
        s = summarizer(i, max_length=256, min_length=64, do_sample=False)
        summarized = summarized + s[0]['summary_text'] +"\n"

    return summarized







import click
from rich.console import Console
from logging import CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG
from transformers.utils import logging
from arxiv_summarizer.summarizer import ArxivSummarizer, SummarizationModel

logging.set_verbosity(ERROR)
logger = logging.get_logger(__name__)

console = Console()

@click.command()
@click.option('--model', default="Falconsai/text_summarization", help='Huggingface id for the model. Defaults to `Falconsai/text_summarization`')
@click.option('--tokenizer', default="Falconsai/text_summarization", help='Huggingface id for the tokenizer.  Defaults to `Falconsai/text_summarization`')
@click.option('--token_length', default=512, help='Maximum token length to use while summarization. Defaults to 512')
@click.argument('arxiv_id')
def main(arxiv_id: str, model: str = "Falconsai/text_summarization", tokenizer: str = "Falconsai/text_summarization", token_length: int = 512):
    summarizer = ArxivSummarizer(
        summarizer = SummarizationModel(
            model = model, 
            tokenizer = tokenizer, 
            
            max_length = token_length, 
            min_length = token_length // 3, 
            do_sample = False,
            truncation = True,
            num_beams = 3,
            early_stopping=True,

            as_cli = True,
        )
    )
    # summarizer = ArxivSummarizer()
    metadata, data = summarizer(arxiv_id)
    
    # print([len(s[0]['summary_text']) for s in data])
    # print(metadata)
    console.print(metadata['Title'], justify = "center")
    # console.print("[ " + metadata['Published'] + " ]", justify = "center")
    console.rule(f"[ {metadata['Published']} ]")
    # console.rule("[bold red]Chapter 2")
    console.print(" ".join([s[0]['summary_text'] for s in data]), justify = "full")

if __name__ == "__main__":
    main()


import torch
from transformers import pipeline

# from tqdm import tqdm

from .utils import doc_loader, strip, chunk
# from .definition_overloading import MultipleMeta

class SummarizationModel:
    """Summarization model for generating summaries.
    """    
    def __init__(self, model, tokenizer, as_cli = False, **kwargs):
        """Initialize a summarization model.

        Args:
            model (str): The pre-trained model name or path.
            tokenizer (str): The pre-trained tokenizer name or path.
            batch_size (int, optional): Batch size for summarization. Defaults to None.
            num_workers (int, optional): Number of workers for summarization. Defaults to None.
            max_length (int, optional): Maximum length of the generated summary. Defaults to None.
            min_length (int, optional): Minimum length of the generated summary. Defaults to None.
            do_sample (bool, optional): Whether to use sampling for summary generation. Defaults to False.
            truncation (bool, optional): Whether to truncate the summary. Defaults to None.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repetition. Defaults to None.
            num_beams (int, optional): Number of beams for beam search. Defaults to 1.
            early_stopping (bool, optional): Whether to stop generation when the stopping criterion is met. Defaults to False.
        """        
        batch_size = kwargs.pop("batch_size", None)
        num_workers = kwargs.pop("num_workers", None)
        piped_model = pipeline("summarization", model=model, tokenizer=tokenizer, batch_size=batch_size, num_workers=num_workers)

        max_length = kwargs.pop("max_length", 10240)
        min_length = kwargs.pop("min_length", 1024)
        do_sample  = kwargs.pop("do_sample", False)

        truncation = kwargs.pop("truncation", None)
        no_repeat_ngram_size = kwargs.pop("no_repeat_ngram_size", None)
        num_beams = kwargs.pop("num_beams", 1)
        early_stopping = kwargs.pop("early_stopping", False)
        self.kwargs = {
                # "max_length": min(max_length, len(x)),            # At most, the size of the input text
                # "min_length": min(min_length, (len(x) * 2) // 3), # At most 66.6% of the size of the input text
                "max_length": max_length,
                "min_length": min_length,
                "no_repeat_ngram_size": no_repeat_ngram_size,
                "num_beams": num_beams,
                "do_sample": do_sample, 
                "truncation": truncation,
                "early_stopping": early_stopping,
        }
        self.as_cli = as_cli

        self.model = [
            lambda x_and_kwargs: piped_model(
                x_and_kwargs[0], 
                **x_and_kwargs[1]
            )
        ]

    def add(self, layer):
        """Add a layer to the summarization model.

        Args:
            layer (callable): A callable layer to be added to the model.
        """        
        self.model.append(layer)

    def __call__(self, text):
        """Generate a summary for the input text using the summarization model.

        Args:
            text (str): The input text to be summarized.

        Returns:
            str: The generated summary.
        """

        result = []
        if self.as_cli:
            from tqdm.rich import tqdm
        else:
            from tqdm import tqdm

        for data in tqdm(text, unit = "sentence"):
            __data__ = data
            for layer in self.model:
                    __data__ = layer(
                            (
                                __data__,
                                {
                                    "max_length": min(self.kwargs["max_length"], len(data)),            # At most, the size of the input text
                                    "min_length": min(self.kwargs["min_length"], (len(data) * 2) // 3), # At most 66.6% of the size of the input text
                                    "do_sample": self.kwargs["do_sample"], 
                                    "truncation": self.kwargs["truncation"],
                                    "no_repeat_ngram_size": self.kwargs["no_repeat_ngram_size"],
                                    "num_beams": self.kwargs["num_beams"],
                                    "early_stopping": self.kwargs["early_stopping"],
                                }
                            )
                    )
            result.append(
                __data__
            )

        return result

class ArxivSummarizer:
    """ArXiv document summarizer.

    Returns:
        str: The generated summary for the ArXiv document.

    Example:
        The `ArxivSummarizer` class can be used with different configurations:

        1. With an existing `SummarizationModel`:

        ```python
        summarizer_model = SummarizationModel(model="Falconsai/text_summarization", tokenizer="Falconsai/text_summarization", max_length=256)
        arxiv_summarizer = ArxivSummarizer(summarizer=summarizer_model)
        summary = arxiv_summarizer(arxiv_id="1234.6789")
        ```

        2. Using a pre-trained model by name:

        ```python
        arxiv_summarizer = ArxivSummarizer(model="facebook/bart-large-cnn", max_length=256)
        summary = arxiv_summarizer(arxiv_id="1234.6789")
        ```

        3. Using default models based on GPU availability:

        ```python
        arxiv_summarizer = ArxivSummarizer()
        summary = arxiv_summarizer(arxiv_id="1234.6789")
        ```
        
    Args:
        summarizer (SummarizationModel): The pre-initialized summarization model.
        model (str, optional): The pre-trained model name or path. Defaults to `Falconsai/text_summarization`_ for CPU and `facebook/bart-large-cnn`_ when CUDA is available.
        max_length (int, optional): Maximum length of the generated summary. Defaults to 256.
        min_length (int, optional): Minimum length of the generated summary. Defaults to 64.
        do_sample (bool, optional): Whether to use sampling for summary generation. Defaults to False.

    Todo:
        * For module TODOs
        * You have to also use ``sphinx.ext.todo`` extension

    .. _Falconsai/text_summarization:
    https://huggingface.co/Falconsai/text_summarization

    .. _facebook/bart-large-cnn:
    https://huggingface.co/facebook/bart-large-cnn
    """

    def __init__(self, summarizer: SummarizationModel = None, model: str = None, max_length: int = 256, min_length: int = 64, do_sample: bool = False):

        if summarizer:
            self.model = summarizer
        else:
            if model is None:
                if torch.cuda.is_available():
                    model = "facebook/bart-large-cnn"
                else:
                    model = "Falconsai/text_summarization"

            self.model = SummarizationModel(
                model=model, 
                tokenizer=model, 
                
                max_length=max_length, 
                min_length=min_length, 
                do_sample=do_sample,
                truncation = True,
                num_beams=3,
                early_stopping=True,
            )

    def __call__(self, arxiv_id=None):
        """Generate a summary for the specified ArXiv document.

        Args:
            arxiv_id (str, optional): The ArXiv document ID. Defaults to None.

        Returns:
            str: The generated summary for the ArXiv document.
        """        
        metadata, page_content = doc_loader(arxiv_id)
        content = strip(page_content)
        sentence = chunk(content)
        return metadata, self.model(sentence)



from setuptools import find_packages, setup

# from sphinx.setup_command import BuildDoc

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="arxiv-summarizer",
    version="0.1.1",
    description="A happy toolkit for arxiv paper summarization and understanding.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ArchanGhosh/Arxiv-Summarizer",
    project_urls={
        "GitHub": "https://github.com/ArchanGhosh/Arxiv-Summarizer",
        "Homepage": "https://github.com/ArchanGhosh/Arxiv-Summarizer",
    },
    author="Archan Ghosh, Arnav Das",
    author_email="gharchan@gmail.com, arnav.das88@gmail.com",
    maintainer="Archan Ghosh, Arnav Das, Debgandhar Ghosh, Subhayu Bala",
    maintainer_email="gharchan@gmail.com, arnav.das88@gmail.com, debgandhar4000@gmail.com, balasubhayu99@gmail.com",
    keywords="arxiv sdk summarization summary",
    license="Apache",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    py_modules=['__main__'],
    entry_points={
        'console_scripts': [
            'arxiv_summarizer = arxiv_summarizer.__main__:cli',
        ],
    },
    # cmdclass={"build_sphinx": BuildDoc},
    # these are optional and override conf.py settings
    command_options={
        "build_sphinx": {
            "project": ("setup.py", "arxiv-summarizer"),
            "version": ("setup.py", "0.1.1a"),
            "release": ("setup.py", "0.1"),
            "source_dir": ("setup.py", "docs"),
        }
    },
    install_requires=[
        # Arxiv & PDF
        "langchain",    # For loading Arxiv PDFs
        "arxiv",        # `langchain` dependency
        "pymupdf",      # `langchain` dependency

        # ML
        "transformers", # For Transformer models
        #"gradio",       # For Web Interface
        "torch==2.1.2", # For Machine Learning

        # CLI
        "click",        # For CLI arguments
        "rich",         # For Rich Text in CLI

        # Tests
        "pytest",       # For Unit Tests
    ],
    extras_require={
        "dev": [
            "nox",
            "pytest",
        ],
        "docs": [
            "sphinx",
            "sphinxemoji",
            "pydata-sphinx-theme",
            "numpydoc",
            "sphinx_panels",
            "matplotlib",
            "Ipython",
            "sphinx-hoverxref",
        ],
    },
    classifiers=[
        # License
        "License :: OSI Approved :: Apache Software License",
        # Project Maturity
        "Development Status :: 3 - Alpha",
        # Topic
        "Topic :: Communications",
        # Intended Audience
        "Intended Audience :: Science/Research",
        # Compatibility
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        #Environment
        "Environment :: Console",
        "Environment :: GPU",
        
        # Python Version
        #"Programming Language :: Python :: 3",
        #"Programming Language :: Python :: 3.4",
        #"Programming Language :: Python :: 3.5",
        #"Programming Language :: Python :: 3.6",
        #"Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)
